@@ -2,7 +2,7 @@
 
 import warnings
 import numpy as np
-from . import viz, regressors, util
+from . import viz, regressors, util, fit, data
 from anamnesis import AbstractAnam, register_class
 
 
@@ -166,11 +166,9 @@ class GLMDesign(AbstractAnam):
     def plot_efficiency(self, showfig=False, savepath=None):
         "Plot the correlation matrix and singular values from the design matrix"
 
-        corr_mat = np.corrcoef(self.design_matrix.T)
-        _, s, _ = np.linalg.svd(self.design_matrix)
-
-        fig = viz.plot_design_efficiency(corr_mat, s, self.num_regressors,
-                                         regressor_names=None, normalise_singlar_values=True)
+        fig = viz.plot_design_efficiency(self.design_matrix,
+                                         regressor_names=self.regressor_names,
+                                         normalise_singlar_values=True)
 
         import matplotlib.pyplot as plt
         if savepath is not None:
@@ -290,8 +288,8 @@ class GLMDesign(AbstractAnam):
         return ret
 
     @classmethod
-    def initialise_from_matrices(cls, design_matrix, contrasts, f_tests=None,
-                                 regressor_names=None, contrast_names=None, f_names=None):
+    def initialise_from_matrices(cls, design_matrix, contrasts, ftests=None,
+                                 regressor_names=None, contrast_names=None, ftest_names=None):
         """
         Initialise class with GLM info and precompute some useful values
         """
@@ -313,15 +311,16 @@ class GLMDesign(AbstractAnam):
             ret.contrast_names = contrast_names
 
         # Define any F tests
-        if f_tests is not None:
-            ret.f_tests = f_tests
-            if f_names is None:
+        if ftests is not None:
+            ret.ftests = ftests
+            if ftest_names is None:
                 # Use default names if they aren't provided
-                ret.f_names = np.arange(ret.f_tests.shape[1]).astype(str)
+                ret.ftest_names = np.arange(ret.ftests.shape[1]).astype(str)
             else:
-                ret.f_names = f_names
+                ret.ftest_names = ftest_names
         else:
-            ret.f_tests = None
+            ret.ftests = None
+            ret.ftest_names = None
 
         ret.sanity_check()
 
@@ -331,6 +330,40 @@ class GLMDesign(AbstractAnam):
 register_class(GLMDesign)
 
 
+def variance_inflation_factor(design_matrix, return_betas=False):
+    """Compute variance inflation factor of each regressor.
+
+    https://en.wikipedia.org/wiki/Variance_inflation_factor
+    """
+
+    num_regressors = design_matrix.shape[1]
+
+    contrasts = np.zeros((1, num_regressors-1))
+    contrast_names = ['A']
+
+    vif = np.zeros((num_regressors,))
+    betas = np.zeros((num_regressors, num_regressors)) * np.nan
+
+    for ind in range(num_regressors):
+        y = data.TrialGLMData(data=design_matrix[:, ind])
+        X = np.delete(design_matrix.copy(), ind, axis=1)
+
+        regressor_names = [str(64 + ii) for ii in range(num_regressors-1)]
+        D = GLMDesign.initialise_from_matrices(X, contrasts,
+                                               contrast_names=contrast_names,
+                                               regressor_names=regressor_names)
+
+        model = fit.OLSModel(D, y)
+        betas[np.setdiff1d(np.arange(num_regressors), ind), ind] = model.betas[:, 0]
+
+        vif[ind] = 1 / (1-model.r_square[0, 0])
+
+    if return_betas:
+        return vif, betas
+    else:
+        return vif
+
+
 class RegressorList(list):
 
     def add(self, rtype, **kwargs):
@@ -412,7 +445,8 @@ class DesignConfig:
 
         # Append to existing contrasts
         for c in self.contrasts:
-            c['values'].append(0)
+            if not isinstance(c['values'], dict):
+                c['values'].append(0)
 
     def add_contrast(self, **kwargs):
         if 'name' not in kwargs:
