@@ -7,7 +7,7 @@ from . import fit
 import sys
 from copy import deepcopy
 
-from scipy import ndimage
+from scipy import ndimage, stats
 
 
 class Permutation:
@@ -110,6 +110,8 @@ class Permutation:
         perm_stat_args={}
         """
 
+        np.random.seed()
+
         if first:
             # Don't permute design matrix
             perm = self._compute_model_stat(self._design, data, fit=fit, smooth_args=smooth_args)
@@ -125,6 +127,8 @@ class Permutation:
         X = self._design.design_matrix.copy()
         if self.perm_type == 'sign-flip':
             perm_inds = np.random.permutation(np.tile([1, -1], int(X.shape[0]/2)))
+            if np.remainder(X.shape[0], 2) == 1:
+                perm_inds = np.r_[perm_inds, np.random.choice([1, -1])]
             X[:, self.cinds] = X[:, self.cinds] * perm_inds[:, None]
         elif self.perm_type == 'row-shuffle':
             perm_inds = np.random.permutation(X.shape[0])
@@ -144,17 +148,17 @@ class Permutation:
                                        smooth_args=smooth_args,
                                        perm_stat_args=kwargs)
 
-        for ii in range(1, self.nperms):
-            nulls[ii, ...] = self._run_perm(data, first=False, fit=fit,
-                                            smooth_args=smooth_args,
-                                            perm_stat_args=kwargs)
+        #for ii in range(1, self.nperms):
+        #    nulls[ii, ...] = self._run_perm(data, first=False, fit=fit,
+        #                                    smooth_args=smooth_args,
+        #                                    perm_stat_args=kwargs)
 
-        #import multiprocessing as mp
-        #p = mp.Pool(processes=self.nprocesses)
-        #args = [(data, False, fit, smooth_args, kwargs) for ii in range(1, self.nperms)]
-        #res = p.starmap(self._run_perm, args)
+        import multiprocessing as mp
+        p = mp.Pool(processes=self.nprocesses)
+        args = [(data, False, fit, smooth_args, kwargs) for ii in range(1, self.nperms)]
+        res = p.starmap(self._run_perm, args)
 
-        #nulls[1:, ...] = np.array(res)
+        nulls[1:, ...] = np.array(res)
 
         self.nulls = nulls
 
@@ -163,7 +167,7 @@ class Permutation:
         if self.nulls is None:
             print('Fit permutations first!')
         else:
-            return np.percentile(self.nulls, percentiles, axis=0)
+            return np.nanpercentile(self.nulls, percentiles, axis=0)
 
     def get_sig_at_percentile(self, percentile):
 
@@ -172,6 +176,25 @@ class Permutation:
         return self.nulls[0, ...] > thresh
 
 
+def permute_design_matrix(design, cinds, perm_type):
+
+    X = design.design_matrix.copy()
+    if perm_type == 'sign-flip':
+        perm_inds = np.random.permutation(np.tile([1, -1], int(X.shape[0]/2)))
+        if np.remainder(X.shape[0], 2) == 1:
+            perm_inds = np.r_[perm_inds, np.random.choice([1, -1])]
+        X[:, cinds] = X[:, cinds] * perm_inds[:, None]
+    elif perm_type == 'row-shuffle':
+        perm_inds = np.random.permutation(X.shape[0])
+        ix = np.ix_(perm_inds, cinds)  # np.ix allows us to apply indexing to both dims
+        X[:, cinds] = X[ix]
+
+    perm_design = deepcopy(design)
+    perm_design.design_matrix = X
+
+    return perm_design
+
+
 class MaxStatPermutation(Permutation):
 
     def _get_null_dims(self, dims, **kwargs):
@@ -385,6 +408,216 @@ def _get_max_cluster_per_dim(cluster_slices, cluster_stats, nonpooled_dims, nonp
     return cluster_stats_nonpooled, cluster_inds_nonpooled
 
 
+def _get_inds(q, ii):
+    """Get location of matching values."""
+    return np.where(q[:, ii])[0]
+
+
+def get_spatial_cluster_inds_from_sensors(x, adj):
+    """Get spatial cluster given data and adjacency matrix."""
+    q = x[:, None].dot(x[None, :]) * adj
+
+    clu = np.zeros_like(x, dtype=int) - 1
+    clust = -1
+    for ii in range(q.shape[0]):
+        # Are there any significant columns.
+        if np.sum(q[:, ii]) > 0:
+            # If so where are they?
+            inds = np.where(q[:, ii])[0]
+
+            # Skip if already a cluster
+            if clu[ii] > -1:
+                continue
+
+            # If not a cluster, iteratively expand inds - 5 steps normally enough...
+            for jj in range(5):
+                inds = np.unique(np.concatenate([_get_inds(q, ii) for ii in inds]))
+
+            clust += 1
+            clu[inds] = clust
+    return clu
+
+
+def shakedown_cluster_numbers(c):
+    cluster_ids, ccounts = np.unique(c, return_counts=True)
+    out = np.zeros_like(c) - 1
+    cnt = 0
+    for idx, idn in enumerate(cluster_ids):
+        if idn == -1:
+            continue  # Drop null region
+        out[c == idn] = cnt
+        cnt += 1
+    return out
+
+
+def get_cluster_extents(x, c, pwr=1):
+    cluster_ids = np.unique(c)
+    out = np.zeros((len(cluster_ids),))
+    for idx, idn in enumerate(cluster_ids):
+        if pwr == 0:
+            out[idx] = np.sum(c == idn)
+        else:
+            out[idx] = np.sum(x[c == idn]**pwr)
+    return cluster_ids, out
+
+
+def get_sensor_time_clusters(X, A, cft=5, pwr=1):
+    """
+    X = [sensors x time]
+    A = modality adjacency
+    """
+
+    # Get connectivity in time for each sensor
+    conn = np.zeros((3, 3))
+    conn[:, 1] = True
+    C_time = ndimage.label(X > cft, conn.T)[0] - 1
+
+    # Get connectivity in space for each timepoint
+    C = np.zeros_like(X)
+    clust = -1
+    for ii in range(X.shape[1]):
+        C[:, ii] = get_spatial_cluster_inds_from_sensors(X[:, ii] > cft, A)
+
+    # Preallocate some stuff
+    C3 = np.zeros_like(C_time) - 1
+    clust = -1
+
+    # for each time_cluster, find all associated space for each sample
+    for ii in range(0, C_time.max()+1):
+        # Find inds of time-cluster
+        a, b = np.where(C_time == ii)
+
+        # Are any of these inds already a cluster? reuse ID if so
+        tmp = np.array([C3[a[idx], b[idx]] for idx in range(len(b))])
+        if np.any(tmp > -1):
+            targ = np.max(tmp)
+        else:
+            clust += 1
+            targ = clust
+
+        for idx in range(len(b)):
+            tmp = C_time[a[idx], b[idx]]
+
+            # original sample is in cluster
+            C3[a[idx], b[idx]] = targ
+
+            # all linked space is in cluster
+            tmp2 = C[a[idx], b[idx]]
+            inds = np.where(C[:, b[idx]] == tmp2)[0]
+            C3[inds, b[idx]] = targ
+    C3 = shakedown_cluster_numbers(C3)
+
+    cinds, cexts = get_cluster_extents(X, C3, pwr=pwr)
+
+    return C3, cinds, cexts
+
+
+class MNEClusterPermutation3(Permutation):
+
+    def _get_null_dims(self, dims, **kwargs):
+        return (self.nperms,)
+
+    def _extract_perm_stat(self, null, **kwargs):
+        from mne.stats.cluster_level import _find_clusters as mne_find_clusters
+
+        threshold = self.perm_args.get('cluster_forming_threshold')
+        adjacency = self.perm_args.get('adjacency', None)
+
+        flatt = null.flatten()
+        clus, cstats = mne_find_clusters(flatt,
+                                         threshold,
+                                         adjacency=adjacency)
+
+        if len(clus) == 0:
+            print('No clusters')
+            return 0
+
+        print('Found {0} clusters - {1} is largest'.format(len(cstats), np.abs(cstats).max()))
+
+        if len(cstats) == 0:
+            return 0
+        else:
+            return np.abs(cstats).max()
+
+    def get_obs_clusters(self, data, fit=fit.OLSModel):
+        from mne.stats.cluster_level import _find_clusters as mne_find_clusters
+        from mne.stats.cluster_level import _reshape_clusters as mne_reshape_clusters
+
+        f = fit(self._design, data)
+        obs = getattr(f, self.perm_metric)[self.contrast_idx, ...]
+
+        threshold = self.perm_args.get('cluster_forming_threshold')
+        adjacency = self.perm_args.get('adjacency')
+
+        flatt = obs.flatten()
+        clus, cstats = mne_find_clusters(flatt,
+                                         threshold,
+                                         adjacency=adjacency)
+
+        clus = mne_reshape_clusters(clus, obs.shape)
+
+        return obs, clus, cstats
+
+    def get_sig_clusters(self, thresh, data):
+        # Find sig clusters
+        obs, clusters, cstats = self.get_obs_clusters(data)
+        thresh = self.get_thresh([thresh])
+        sigs = np.abs(cstats) > thresh
+        sig_inds = np.where(sigs)[0]
+
+        # Collate info from sig clusters
+        out = []
+        for idx in sig_inds:
+            clu = clusters[idx]
+            pval = stats.percentileofscore(self.nulls, cstats[idx])
+            out.append((cstats[idx], pval, clu))
+
+        # Sort from largest to smallest
+        I = np.argsort([c[0] for c in out])[::-1]
+        out = [out[ii] for ii in I]
+
+        return out, obs
+
+
+class MNEClusterPermutation2(Permutation):
+
+    def _get_null_dims(self, dims, **kwargs):
+        return (self.nperms,)
+
+    def _extract_perm_stat(self, null, **kwargs):
+
+        threshold = self.perm_args.get('cluster_forming_threshold')
+        connectivity = self.perm_args.get('connectivity')
+
+        clusters, cinds, cstats = get_sensor_time_clusters(null, connectivity, cft=threshold, pwr=1)
+
+        if (len(cstats) == 1) and (cinds[0] == -1):
+            print('No clusters')
+            return 0
+        else:
+            goods = np.where(cinds > -1)[0]
+            cstats = cstats[goods]
+
+        print('Found {0} clusters - {1} is largest'.format(len(cstats), cstats.max()))
+
+        if len(cstats) == 0:
+            return 0
+        else:
+            return cstats.max()
+
+    def get_obs_clusters(self, data, fit=fit.OLSModel):
+
+        f = fit(self._design, data)
+        obs = getattr(f, self.perm_metric)[self.contrast_idx, ...]
+
+        threshold = self.perm_args.get('cluster_forming_threshold')
+        connectivity = self.perm_args.get('connectivity')
+
+        clusters, cinds, cstats = get_sensor_time_clusters(obs, connectivity, cft=threshold)
+
+        return obs, clusters, cstats
+
+
 class MNEClusterPermutation(Permutation):
 
     def _get_null_dims(self, dims, **kwargs):
@@ -429,6 +662,9 @@ class MNEClusterPermutation(Permutation):
         return clusters, cluster_masks, cstat
 
 
+# -----------------------------------------------
+
+
 def permute_glm(glmdes, data, nperms=5000, stat='cope',
                 maxstat_axes=None, stat_corr_mode=None, cluster_forming_threshold=None,
                 temporal_varcope_smoothing=None, nprocesses=1, smooth_dims=None):
